
/*version for sp2*/

/* 
   OM3:  Copyright 1996-1999 John Anderson
   from om3v1_5.cx with minor bug fixes for multiproc restart consistancy

   om3_v1.4.cx -> om3.c
   128 x 128 point model on pole to pole geometry, currently with reduced
   metric for latitudes > 70 deg.

   Hydrostatic level model with 16 levels taken as the odd levels
   from the GFDL 33 level set.

   Double time split (LF advection/diffustion, F-B gravity, Euler external)
   with reduced surface gravity.
*/

#define PREFIX "./"


/*
   Define file names for restart init and save files if needed.
   This should become part of an integrated model control procedure
   ASAP.
*/
/* #define RestartRun */
/* #define RestartSave */
#define RestartRunName "om3.restart"
#define RestartSaveName "om3.save"


#define Nrows 4
#define Ncols 2


#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "mpi.h"

#define Np (Ncols*Nrows)
#define Lx 4.0e7  /* Global domain from 0W - 360W */
#define Ly 2.0e7  /* 90S - 90N */
#define Nx 128
#define Ny 128
#define Nz 16
#define Mx ((Nx/Ncols)+2)
#define My ((Ny/Nrows)+2)
#define Mz Nz

#define	N_Days	480.0
#define Delx (Lx / Nx)
#define Dely (Ly / Ny)
#define Diff_Mks 8.0e3
#define Diff_Ratio 0.5
#define Restore (1.0/(N_Days*24.0*3600.0))
#define G 0.50
#define N_Split 3
#define Sub_Split 8

#define Delt (6.0*3600.0)
#define N_Steps ((int)((N_Days*24.0*3600.0/Delt)+0.499))
#define N_Blocks 1

int num_nodes, my_address;
int row,col,north,south,east,west;
int ofile;

float uvts_levels[Mz],w_levels[Mz]; /* read from geometry file */
float h[Mz][My][Mx], uvh[Mz][My][Mx]; /* level thickneses, 0 if no water */

float strx[My][Mx],stry[My][Mx]; /* wind stress */ 
float annual_salt[Mz][My][Mx],annual_temp[Mz][My][Mx]; /* mean salt and temp for spin up */

float u[Mz][My][Mx],ul[Mz][My][Mx],ux[Mz][My][Mx],dtu[Mz][My][Mx]; /* State Vars */
float v[Mz][My][Mx],vl[Mz][My][Mx],vx[Mz][My][Mx],dtv[Mz][My][Mx];
float t[Mz][My][Mx],tl[Mz][My][Mx],tx[Mz][My][Mx],dtt[Mz][My][Mx];
float s[Mz][My][Mx],sl[Mz][My][Mx],sx[Mz][My][Mx],dts[Mz][My][Mx];
float ptop[My][Mx],ptopl[My][Mx],ptopx[My][Mx]; /* surface pressure */
/* end of state */

float mix_u[Mz][My][Mx],mix_v[Mz][My][Mx],mix_t[Mz][My][Mx],mix_s[Mz][My][Mx];
float risq[Mz][My][Mx], nsq[Mz][My][Mx], kuv[Mz][My][Mx], kts[Mz][My][Mx];
/* Richarson number squared, k's */

float w[Mz][My][Mx],p[Mz][My][Mx],dzt[Mz][My][Mx],dzs[Mz][My][Mx]; /* derived state variables */
float dzu[Mz][My][Mx],dzv[Mz][My][Mx]; /* derived state variables */
float ubaro[My][Mx],vbaro[My][Mx],dtubaro[My][Mx],dtvbaro[My][Mx]; /* barotropic split variables */
float ubstar[My][Mx],vbstar[My][Mx],ptstar[My][Mx]; /* Euler step temps */

float f[My],taux[My][Mx],tauy[My][Mx]; /* Derived geometries */
float geo1[My], geo2[My], geo3[My];
float depth[My][Mx], uvdepth[My][Mx], rdepth[My][Mx], ruvdepth[My][Mx];

float base_t[Mz],base_s[Mz]; /* base stratification */

double run_time; /* forcing history variables */
float t_forcing[My][Mx],s_forcing[My][Mx];

float work[My][Mx];
char buff[256];
char srbuff[Mx*My*8*10];
FILE *of, *hfile;

MPI_Datatype mpi_edge_type;


unsigned char picbuf[My-2][Mx-2];

void init_state(); /* initialization routines */
void read_geometry();
void read_stress();
void read_annual_temp();
void read_annual_salt();

void time_step(); /* model run time code */
float density(float s, float t, int iz);
void filt4s(float q[Mz][My][Mx]);
void filt4m(float q[Mz][My][Mx]);
void filt4p(float q[My][Mx]);
void wrap_q(float a[My][Mx]);
void wrap_q3(float a[My][Mx], float b[My][Mx], float c[My][Mx]);
void wrap_qz(float a[Mz][My][Mx]);

void write_picture(float a[My][Mx],float vmax); /* movie and history file output
routines */
void fwrite_picture(float a[My][Mx],float vmax);
void write_history(float a[Mz][My][Mx]);
void write_history_slice(float a[Mz][My][Mx], int iz);

float f_max(float a[My][Mx]); /* non-timestep utility routines */
void load_restart();
void read_restart_slice(float a[My][Mx], FILE *tfile);
void save_restart();
void write_restart_slice(float a[My][Mx], FILE *tfile);
void global_write();


main(int argc, char *argv[])
{
   int iblock,istep,ix,iy;
   float max_u, max_v, max_t, max_p;
   float max_an_t,max_an_s, max_sx, max_sy;
   float max_tf,max_sf;
   int ellen,elcnt,stride;

   MPI_Init(&argc,&argv);
   MPI_Buffer_attach(srbuff, Mx*My*8*10);

   ellen = sizeof(float); /* make mti type for e-w edge */
   elcnt = My;
   stride = Mx * ellen;
   MPI_Type_vector(elcnt, ellen, stride, MPI_CHAR, &mpi_edge_type);
   MPI_Type_commit(&mpi_edge_type);

   /* Get self address and number of nodes.*/

   MPI_Comm_rank(MPI_COMM_WORLD, &my_address);
   MPI_Comm_size(MPI_COMM_WORLD, &num_nodes);

   /* open stdout file on sphome for root node, flush on other nodes */

   printf("address = %d of %d\n", my_address, num_nodes);

   if(my_address <= 0 )
   {
     sprintf(buff,"%sstdout_%d_%d.%d",PREFIX,Nrows,Ncols,my_address);

     /*     if((of = freopen(buff, "w", stdout)) == NULL)
       { 
	 printf("of = freopen error");
	 exit(89);
       }
       */
   }
   /*
   else
   {
     if((of = freopen("/dev/null", "w", stdout)) == NULL)
       {
	 printf("second of = freopen error\n");
	 exit(89);
       }
   }
   */

   /* open picture file */

   if(my_address == 0) {
       sprintf(buff,"%som3_%d_%d.cmpics",PREFIX,Nrows,Ncols);
       ofile = global_open(buff, O_CREAT | O_WRONLY, 0666);
   }



   /* open history file */

   if(my_address == 0) {
      sprintf(buff,"%som3_%d_%d.cmhist",PREFIX,Nrows,Ncols);
      hfile = fopen(buff,"wb");
      fprintf(hfile,"%d\n",N_Blocks);
   }



   /* validate machine geometry */

   if(Np != num_nodes) {
     printf("Node number mismatch - needed %d got %d ...exiting\n",Np,num_nodes);
     exit(89);
   }
   else
     printf("Started on %d nodes.\n",num_nodes);
   fflush(of);


   /* Calculate node geometery and neighbor node numbers */

   if(my_address == 0)
      printf("Ncols,Nrows %d %d Mx,My: %d %d\n",Ncols,Nrows,Mx,My);

   row = my_address / Ncols;
   col = my_address % Ncols;

   if(col == 0) west = my_address + (Ncols-1); /* 2-D torus */
   else west = my_address - 1;

   if(col == Ncols-1) east = my_address - (Ncols-1);
   else east = my_address + 1;

   if(row == 0) south = my_address + (Nrows-1)*Ncols;
   else south = my_address - Ncols;

   if(row == Nrows-1)  north = my_address - (Nrows-1)*Ncols;
   else north = my_address + Ncols;

   /* Initialization processing */


   read_geometry();
   read_stress();
   read_annual_temp();
   read_annual_salt();


   write_picture(strx,0.1);
   write_picture(stry,0.1);
   write_picture(&(annual_temp[Nz-1][0]),10.0);
   write_picture(&(annual_salt[Nz-1][0]),2.0);

   max_sx = f_max(strx);
   max_sy = f_max(stry);
   max_an_t = f_max(&(annual_temp[Nz-1][0]));
   max_an_s = f_max(&(annual_salt[Nz-1][0]));

   if(my_address == 0)
   {
	 printf("Max sx: %f, sy: %f, ann. temp, %f, ann. salt %f\n",
            max_sx,max_sy,max_an_t,max_an_s);
   }
   fflush(of);


   init_state();



#ifdef RestartRun
   load_restart();
   if(my_address == 0)
   {
     printf("Model initialized from restart file: %s\n",RestartRunName);
   }
   fflush(of);
#endif

   max_u = f_max(&(u[Nz-1][0]));
   max_v = f_max(&(v[Nz-1][0]));
   max_t = f_max(&(t[Nz-1][0]));
   max_p = f_max(&(ptop[0]));

   if(my_address == 0)
   {
      printf("istep: %d Max u: %f, v: %f, t: %f, p: %f\n",istep,max_u,max_v,max_t,max_p);
      printf("Delt = %f seconds, Block time = %f days.\n",Delt,Delt*N_Steps/(3600.0*24.0));
   }
   fflush(of);

   for(iblock = 0; iblock < N_Blocks; iblock++)
   {

      printf("****** START BLOCK %d of %d *********************\n",
	     iblock+1,N_Blocks);
      fflush(of);


      for(istep = 0; istep < N_Steps; istep++)
      {
   if(my_address == 0)
            printf("Timestep %d\n",istep);
	    time_step();
 	    fflush(of);
      }
      run_time += Delt*N_Steps;


      max_u = f_max(&(u[Nz-1][0]));
      max_v = f_max(&(v[Nz-1][0]));
      max_t = f_max(&(t[Nz-1][0]));
      max_p = f_max(&(ptop[0]));

      if(my_address == 0)
      {
	    printf("Max u: %f, v: %f, t: %f, p: %f\n",max_u,max_v,max_t,max_p);
      }

      max_tf = f_max(&(t_forcing[0]));
      max_sf = f_max(&(s_forcing[0]));

      if(my_address == 0)
      {
	    printf("Max tf: %e, sf: %e\n",max_tf/run_time,max_sf/run_time);
      }

      fflush(of);

      if(iblock >= 0)
      {
        write_picture(&(t_forcing[0]),fabs(max_tf)/3.0);
        write_picture(&(s_forcing[0]),fabs(max_sf)/3.0);
      }

      if(iblock >= 0)
      {
        write_history(u);
        write_history(v);
        write_history(w);
        write_history(t);
        write_history(s);
        write_history(p);
        write_history(risq);

      }
   }
   if(my_address==0)
   {
      close(ofile);
      fclose(hfile);
   }
#ifdef RestartSave
   save_restart();
   if(my_address == 0)
   {
       printf("Model state saved to restart file: %s\n",RestartSaveName);
   }
   fflush(of);
#endif

   MPI_Finalize();  /* clean up world */

}

float f_max(a)
float a[My][Mx];
{
   int ix,iy,maxix,maxiy;
   float max,gmax,gmin,maxamp;

   maxamp = -1.0;

   for(iy = 1; iy < My-1; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++)
      {
	     if((fabs(a[iy][ix]) > maxamp) &&(h[Nz-1][iy][ix+1] > 1.0))
         {
             max = a[iy][ix];
             maxamp = fabs(a[iy][ix]);
             maxix = ix;
             maxiy = iy;
         }
      }
   }
/*
   if(my_address == 0) printf("Max ix,iy: %d %d %f\n",maxix,maxiy,max);
*/

   MPI_Allreduce(&max, &gmax, 1, MPI_FLOAT, MPI_MAX, MPI_COMM_WORLD);
   MPI_Allreduce(&max, &gmin, 1, MPI_FLOAT, MPI_MIN, MPI_COMM_WORLD);



   if(gmax > -gmin) return gmax;
   else return gmin;
}

float f_total(a)
float a[My][Mx];
{
   int ix,iy;
   float total,gtotal;

   total = 0.0;

   for(iy = 1; iy < My-1; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++)
      {
         {
            if(h[Nz-1][iy][ix+1] > 1.0)
              total = total + a[iy][ix];
         }
      }
   }
   MPI_Allreduce(&total,&gtotal,1,MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);


   return gtotal;
}

/*
 * We write a picture one node at a time in cm5 order.
 * In this (non piofs) version each write is staged throught the
 * root node.
 */
void write_picture(float a[My][Mx],float vmax)
{
   int ix,iy;
   int len,pixel_val;
   float val,scl;
   char *bptr,buf[200];

   fflush(of);

   /* write header with Nx, Ny, Ncols, Nrows */

   sprintf(buf,"%d %d %d %d\n",Nx,Ny,Ncols,Nrows);
   len = strlen(buf);
   if(my_address == 0) write(ofile,buf,len);

   scl = 40.0/(vmax);

   for(iy = 1; iy < My-1; iy++)
   {
     for(ix = 1; ix < Mx-1; ix++) /* fill buffer in parallel */
     {
	   val = scl * a[iy][ix];
	   pixel_val = ((int) fabs(val+0.5)) % 40;

	   if(val > 0)
         pixel_val = 41 + pixel_val; /* red colors in std map */
	   else
         pixel_val = 81 + pixel_val; /* blue colrs in std map */

	   if(h[Nz-1][iy][ix] > 1.0)
         picbuf[iy-1][ix-1] = pixel_val;
       else
         picbuf[iy-1][ix-1] = 0; /* zero out dry cells */
    }
  }
  global_write();
}

void fwrite_picture(float a[My][Mx],float vmax)
{
   int ix,iy;
   int len,pixel_val;
   float val,scl;
   char *bptr,buf[200];

   /* write header with Nx, Ny, Ncols, Nrows */
   sprintf(buf,"%d %d %d %d\n",Nx,Ny,Ncols,Nrows);
   len = strlen(buf);
   if(my_address == 0) write(ofile,buf,len);

   scl = 40.0/(vmax);

   for(iy = 1; iy < My; iy++)
   {
      for(ix = 1; ix < Mx-1; ix++) /* fill buffer in parallel */
      {
        if(h[Nz-1][iy][ix+1] > 1.0)
	      work[iy][ix] = 0.5 * (a[iy][ix] + a[iy][ix+1]);
        else
          work[iy][ix] = a[iy][ix];
      }
   }

   for(iy = 1; iy < My-1; iy++)
   {
     for(ix = 1; ix < Mx-1; ix++) /* fill buffer in parallel */
     {
       if(h[Nz-1][iy+1][ix] > 1.0)
	     val = scl * 0.5 * (work[iy][ix] + work[iy+1][ix]);
       else
         val = scl * work[iy][ix];
	 
	   pixel_val = ((int) fabs(val+0.5)) % 40;

	   if(val > 0)
         pixel_val = 41 + pixel_val; /* red colors in std map */
	   else
         pixel_val = 81 + pixel_val; /* blue colrs in std map */

	   if(h[Nz-1][iy][ix] > 1.0)
         picbuf[iy-1][ix-1] = pixel_val;
       else
         picbuf[iy-1][ix-1] = 0; /* zero out dry cells */
     }
  }
  global_write();
}

void global_write()
{
   int i, j, k, l, ik, jl, ierr;
   int tt;
   unsigned char t1, t2;
   MPI_Status status;


   ierr = MPI_Bsend(picbuf, (Mx-2)*(My-2), MPI_CHAR, 0, 99, MPI_COMM_WORLD);
   if(ierr != 0) {
     printf("Bsend error %d\n", ierr);
     exit(1);
   }


   if(my_address == 0) {
       for(i=0; i<Nrows; i++)
       {
	     for(j=0; j<Ncols; j++)
         {
 
            MPI_Recv(picbuf, (Mx-2)*(My-2), MPI_CHAR, i*Ncols+j,
		      99, MPI_COMM_WORLD, &status);

       	    write(ofile, picbuf, (Mx-2)*(My-2));
	     }
       }
   }
}

/* History write routines */

void write_history(float a[Mz][My][Mx])
{
  int iz;

  for(iz = 0; iz < Mz; iz++)
  {
     write_history_slice(a,iz);
  }
}

unsigned char hist_out[My][Mx][2];

void write_history_slice(float a[Mz][My][Mx], int iz)
{
  int ix,iy,ival,ic;
  float xx,lxmin,lxmax,xmin,xmax,scale;
  int jrow,jcol,jy;

  unsigned char row_out[Mx-2][2];
  MPI_Status status;
  int ierr;


  lxmin = 1.0e10;
  lxmax = -1.0e10;

  for(iy = 1; iy < My-1; iy++)
  {
    for (ix = 1; ix < Mx-1; ix++)
    {
      if(h[iz][iy][ix] > 1.0)
      {
         xx = a[iz][iy][ix];
         if(xx < lxmin) lxmin = xx;
         if(xx > lxmax) lxmax = xx;
      }
    }
  }


   MPI_Allreduce(&lxmax, &xmax, 1, MPI_FLOAT, MPI_MAX, MPI_COMM_WORLD);
   MPI_Allreduce(&lxmin, &xmin, 1, MPI_FLOAT, MPI_MIN, MPI_COMM_WORLD);



  if(my_address == 0)
  {
     fprintf(hfile,"%f %f\n",xmin,xmax);
  }
  xx = xmax-xmin;
  if(xx < 1.0e-10) xx=1.0e-10;
  scale = 60000.0/xx;

  for(iy = 1; iy < My-1; iy++)
  {
    for (ix = 1; ix < Mx-1; ix++)
    {
      if(h[iz][iy][ix] > 1.0)
      {
         xx = a[iz][iy][ix];
         ival = 100 + scale*(xx-xmin);
      }
      else ival = 0;

      hist_out[iy][ix][0] = ival/256;
      hist_out[iy][ix][1] = ival % 256;
    }
  }
   for(jy = 0; jy < Ny; jy++) /* write xy field in natural order*/
   {
     jrow = jy/(My-2);
     iy = jy - jrow*(My-2) + 1;
     for(jcol = 0; jcol < Ncols; jcol++) /* write in column order */
     {
        if((col == jcol) && (row == jrow)) /* Select source processor */
        {
           ierr = MPI_Bsend(&(hist_out[iy][1][0]), 2*(Mx-2),
                    MPI_CHAR, 0, 99, MPI_COMM_WORLD);
           if(ierr != 0)
           {
              printf("Bsend error in write_history_slice: %d\n", ierr);
              fflush(of);
              exit(1);
           }
        }
        if(my_address == 0)
        {
           MPI_Recv(row_out, 2*(Mx-2), MPI_CHAR, jrow*Ncols+jcol,
		      99, MPI_COMM_WORLD, &status);
           fwrite(row_out,sizeof(char),2*(Mx-2),hfile);
        }
        MPI_Barrier( MPI_COMM_WORLD);

     }
   }
}

void read_geometry()
{
   int ix,iy,jx,jy,jz;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",PREFIX,"om3.geometry");
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on geometry failed.\n");
      exit(89);
   }

   /* read depths of two types of grid levels */

   if(my_address == 0) fread(in,sizeof(char),2*Nz,tfile);
   MPI_Bcast(in,2*Nz,MPI_CHAR,0,MPI_COMM_WORLD);
   inptr = in;
   for(jz = 0; jz < Nz; jz++)
   {
      t1 = *inptr++; t2 = *inptr++;
      tt = t1*256 + t2;
      uvts_levels[jz] = tt;
      /*printf("%d %f\n",jz,uvts_levels[jz]);*/
   }

   if(my_address == 0) fread(in,sizeof(char),2*Nz,tfile);
   MPI_Bcast(in,2*Nz,MPI_CHAR,0,MPI_COMM_WORLD);
   inptr = in;
   for(jz = 0; jz < Nz; jz++)
   {
      t1 = *inptr++; t2 = *inptr++;
      tt = t1*256 + t2;
      w_levels[jz] = tt;
   }

   for(jz = Nz-1; jz >= 0; jz--) /* read 3-D topo field */
   {
     for(jy = 0; jy < Ny; jy++)
     {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0) fread(in,sizeof(char),2*Nx,tfile);
      MPI_Bcast(in,2*Nx,MPI_CHAR,0,MPI_COMM_WORLD);
      inptr = in;
      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
	     t1 = *inptr++; t2 = *inptr++;
         tt = t1*256 + t2;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
         {
            h[jz][iy][ix] = (float)tt;
         }
      }
     }
   }

   if(my_address == 0) fclose(tfile);

   wrap_qz(h);

   for(iy = 1; iy < My-1; iy++) /* fill depth and reciprical depth arrays */
   {
     for(ix = 1; ix < Mx-1; ix++)
     {
       depth[iy][ix] = 0.0;
       uvdepth[iy][ix] = 0.0;
       for(jz = 0; jz < Nz; jz++)
       {
          depth[iy][ix] += h[jz][iy][ix];

	      if((h[jz][iy][ix+1] > 1.0) && (h[jz][iy][ix-1] > 1.0)
          &&(h[jz][iy+1][ix] > 1.0) && (h[jz][iy-1][ix] > 1.0))
          {
             uvdepth[iy][ix] += h[jz][iy][ix];
             uvh[jz][iy][ix] = h[jz][iy][ix];
          }
          else
          {
             uvh[jz][iy][ix] = 0.0;
          }
       }
       if(depth[iy][ix] > 1.0)
         rdepth[iy][ix] = 1.0/depth[iy][ix];
       else
         rdepth[iy][ix] = 0.0;

       if(uvdepth[iy][ix] > 1.0)
         ruvdepth[iy][ix] = 1.0/uvdepth[iy][ix];
       else
         ruvdepth[iy][ix] = 0.0;
     }
   }
   wrap_q(depth);
   wrap_q(uvdepth);
   wrap_q(rdepth);
   wrap_q(ruvdepth);
   wrap_qz(uvh);
}

void read_stress()
{
   int ix,iy,jx,jy;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",PREFIX,"om3.stress");
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on stress failed.\n");
      exit(89);
   }


   for(jy = 0; jy < Ny; jy++) /* read x-stress field */
   {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0) fread(in,sizeof(char),2*Nx,tfile);
      MPI_Bcast(in,2*Nx,MPI_CHAR,0,MPI_COMM_WORLD);
      inptr = in;

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
         t1 = *inptr++; t2 = *inptr++;
         tt = t1*256 + t2;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
            strx[iy][ix] = ((float)tt/10000.0)-0.3;
      }
   }

   for(jy = 0; jy < Ny; jy++) /* read y-stress field */
   {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0) fread(in,sizeof(char),2*Nx,tfile);
      MPI_Bcast(in,2*Nx,MPI_CHAR,0,MPI_COMM_WORLD);
      inptr = in;

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
	     t1 = *inptr++; t2 = *inptr++;
         tt = t1*256 + t2;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
            stry[iy][ix] = ((float)tt/10000.0)-0.3;
      }
   }
   if(my_address == 0) fclose(tfile);
}

void read_annual_temp()
{
   int ix,iy,jx,jy,jz;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;
   float flin[Nx];

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",PREFIX,"om3.temp");
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on temp failed.\n");
      exit(89);
   }

   for(jz = Nz-1; jz >= 0; jz--) /* read 3-D temp field */
   {
     for(jy = 0; jy < Ny; jy++)
     {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0)
      { float fxx;
        fread(in,sizeof(char),2*Nx,tfile);
        inptr = in;
        fxx=0.0;
        for(jx = 0; jx < Nx; jx++)
        {
          t1 = *inptr++; t2 = *inptr++;
          tt  = t1*256 + t2;
          flin[jx] =((float)tt-10000.0)/100.0;
          fxx += flin[jx];
        }
        fxx = fxx/Nx;
        for(jx = 0; jx < Nx; jx++) flin[jx]=fxx;
        if(jy == 85) base_t[jz] = flin[0];
      }

    MPI_Bcast(flin,Nx,MPI_FLOAT,0,MPI_COMM_WORLD);

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
         {
            annual_temp[jz][iy][ix] = flin[jx];
         }
      }
     }
   }
   if(my_address == 0) fclose(tfile);
 MPI_Bcast(base_t,Mz,MPI_FLOAT,0,MPI_COMM_WORLD);
   wrap_qz(annual_temp);
}

void read_annual_salt()
{
   int ix,iy,jx,jy,jz;
   FILE *tfile;
   int tt;
   unsigned char in[2*Nx], *inptr, t1, t2;
   float flin[Nx];

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",PREFIX,"om3.salt");
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on temp failed.\n");
      exit(89);
   }

   for(jz = Nz-1; jz >= 0; jz--) /* read 3-D salinity field */
   {
     for(jy = 0; jy < Ny; jy++)
     {
      iy = jy - row*(My-2) + 1;
      if(my_address == 0)
      { float fxx;
        fread(in,sizeof(char),2*Nx,tfile);
        inptr = in;
        fxx=0.0;
        for(jx = 0; jx < Nx; jx++)
        {
          t1 = *inptr++; t2 = *inptr++;
          tt  = t1*256 + t2;
          flin[jx] =((float)tt-10000.0)/100.0;
          fxx += flin[jx];
        }
        fxx = fxx/Nx;
        for(jx = 0; jx < Nx; jx++) flin[jx]=fxx;
        if(jy == 85) base_s[jz] = flin[0];
      }

    MPI_Bcast(flin,Nx,MPI_FLOAT,0,MPI_COMM_WORLD);

      for(jx = 0; jx < Nx; jx++)
      {
         ix = jx - col*(Mx-2) + 1;
         if((ix >= 0) && (ix < Mx) && (iy >= 0) && (iy < My))
         {
            annual_salt[jz][iy][ix] = flin[jx];
         }
      }
     }
   }
   if(my_address == 0) fclose(tfile);
 MPI_Bcast(base_s,Mz,MPI_FLOAT,0,MPI_COMM_WORLD);
   wrap_qz(annual_salt);
}

void init_state()
{
   int ix,iy,iz,jx,jy,mx,my;
   float x,y,yeq,lat,omega;

   omega = (366.0/365.0) * 2.0*3.14159 / (3600.0*24.0);
   run_time = 0.0;

    for (iy = 0; iy < My; iy++) /* convective adjustment */
    {
      for (ix = 0; ix < Mx; ix++)
      { int mixing;

        mixing = 1;
        while(mixing)
        {
          mixing = 0;
	      for (iz = Mz-2; iz >= 0; iz--)
	      { float d1 ,d2 , snew, tnew;
            if((h[iz][iy][ix] > 1.0) && (h[iz+1][iy][ix]) > 1.0)
            {
               d1 = density(annual_salt[iz][iy][ix],annual_temp[iz][iy][ix],iz);
               d2 = density(annual_salt[iz+1][iy][ix],annual_temp[iz+1][iy][ix],iz);
               if(d2 > (d1 + 1.0e-7))
               {
                  mixing = 1;

                  snew = (annual_salt[iz][iy][ix]*h[iz][iy][ix] +
                          annual_salt[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);
                  annual_salt[iz][iy][ix] = annual_salt[iz+1][iy][ix] = snew;

                  tnew = (annual_temp[iz][iy][ix]*h[iz][iy][ix] +
                          annual_temp[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);
                  annual_temp[iz][iy][ix] = annual_temp[iz+1][iy][ix] = tnew;
               }
            }
	      }
	    }
      }
    }


  for(iz = 0; iz < Mz; iz++)
  {
    for(iy = 0; iy < My; iy++)
    {
      for(ix = 0; ix < Mx; ix++)
      {
	    ul[iz][iy][ix] = u[iz][iy][ix] = ux[iz][iy][ix] = 0.0;
	    vl[iz][iy][ix] = v[iz][iy][ix] = vx[iz][iy][ix] = 0.0;
	    tl[iz][iy][ix] = t[iz][iy][ix] = tx[iz][iy][ix] = base_t[iz];
	    sl[iz][iy][ix] = s[iz][iy][ix] = sx[iz][iy][ix] = base_s[iz];
       }
     }
  }

  for(iy = 0; iy < My; iy++)
  {
    for(ix = 0; ix < Mx; ix++)
    {
	   ptopl[iy][ix] = ptop[iy][ix] = ptopx[iy][ix] = 0.0;
       t_forcing[iy][ix] = 0.0;
       s_forcing[iy][ix] = 0.0;
     }
   }

   for(iy = 0; iy < My; iy++)
   {  float x1,x2,x3,del_lat;


      jy = row*(My-2) + (iy-1);
      yeq = Dely * (Ny/2.0);
      y = Dely * (jy+0.5);
      lat = (2.0*3.14159/4.0e7) * (y-yeq);
      del_lat = 3.14159/Ny;
      f[iy] = 2.0*omega*sin(lat);

      x1 = cos(lat-del_lat);
      x2 = cos(lat);
      x3 = cos(lat+del_lat);

      geo1[iy] = x1/x2;
      geo2[iy] = x3/x2;

      geo3[iy]=1.0/cos(lat);
      if(geo3[iy] > 2.0) geo3[iy]=2.0;

      for(ix = 0; ix < Mx; ix++)
      {
	    if(h[Nz-1][iy][ix] > 1.0)
	    {
	       taux[iy][ix] = strx[iy][ix]/(1030.0 * h[Nz-1][iy][ix]);
	       tauy[iy][ix] = stry[iy][ix]/(1030.0 * h[Nz-1][iy][ix]);
	    }
	    else
	    {
	       taux[iy][ix] = 0.0;
	       tauy[iy][ix] = 0.0;
	    }
      }
   }
   wrap_q(taux);
   wrap_q(tauy);
}

void load_restart()
{
   int jz;
   FILE *tfile;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",PREFIX,RestartRunName);
      tfile = fopen(buff,"r");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on restart read failed.\n");
      fflush(of);
      exit(89);
   }

   for(jz = 0; jz < Nz; jz++) /* model state fields */
   {
      read_restart_slice(&(u[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(ul[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(v[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(vl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(t[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(tl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(s[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      read_restart_slice(&(sl[jz][0]),tfile);
   }
   read_restart_slice(&(ptop[0]),tfile);
   read_restart_slice(&(ptopl[0]),tfile);

   if(my_address == 0)
   {
      fread(&run_time,sizeof(double),1,tfile);
   }
   read_restart_slice(&(t_forcing[0]),tfile);
   read_restart_slice(&(s_forcing[0]),tfile);

   if(my_address == 0) fclose(tfile);
}

void read_restart_slice(float a[My][Mx], FILE *tfile)
{
   int ix,iy,jx,jy;
   float row_in[Nx];

   for(jy = 0; jy < Ny; jy++) /* read xy field */
   {
     iy = jy - row*(My-2) + 1;
     if(my_address == 0) fread(row_in,sizeof(float),Nx,tfile);
     MPI_Bcast(row_in,Nx,MPI_FLOAT,0,MPI_COMM_WORLD);
     for(jx = 0; jx < Nx; jx++)
     {
       ix = jx - col*(Mx-2) + 1;
       if((ix > 0) && (ix < Mx-1) && (iy > 0) && (iy < My-1))
       {
          a[iy][ix] = row_in[jx];
       }
     }
   }
   wrap_q(a);
}

void save_restart()
{
   int jz;
   FILE *tfile;

   if(my_address == 0)
   {
      sprintf(buff,"%s%s",PREFIX,RestartSaveName);
      tfile = fopen(buff,"w");
   }
   if((tfile == 0) && (my_address == 0))
   {
      printf("fopen on restart write failed.\n");
      fflush(of);
      exit(1);
   }
   else if (my_address == 0)
   {
      printf("open ok \n");
      fflush(of);
   }

   for(jz = 0; jz < Nz; jz++) /* model state fields */
   {
      write_restart_slice(&(u[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(ul[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(v[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(vl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(t[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(tl[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(s[jz][0]),tfile);
   }
   for(jz = 0; jz < Nz; jz++)
   {
      write_restart_slice(&(sl[jz][0]),tfile);
   }
   write_restart_slice(&(ptop[0]),tfile);
   write_restart_slice(&(ptopl[0]),tfile);

   if(my_address == 0)
   {
      fwrite(&run_time,sizeof(double),1,tfile);
   }
   write_restart_slice(&(t_forcing[0]),tfile);
   write_restart_slice(&(s_forcing[0]),tfile);

   if(my_address == 0) fclose(tfile);
}

void write_restart_slice(float a[My][Mx], FILE *tfile)
{
   int jrow,jcol,jy,iy;

   float row_out[Mx-2];
   MPI_Status status;
   int ierr;


   for(jy = 0; jy < Ny; jy++) /* write xy field in natural order*/
   {
     jrow = jy/(My-2);
     iy = jy - jrow*(My-2) + 1;
     for(jcol = 0; jcol < Ncols; jcol++) /* write in column order */
     {
        if((col == jcol) && (row == jrow)) /* Select source processor */
        {
           ierr = MPI_Bsend(&(a[iy][1]), Mx-2, MPI_FLOAT, 0, 99, MPI_COMM_WORLD);
           if(ierr != 0)
           {
              printf("Bsend error in write_restart_slice: %d\n", ierr);
              fflush(of);
              exit(1);
           }
        }
        if(my_address == 0)
        {
           MPI_Recv(row_out, Mx-2, MPI_FLOAT, jrow*Ncols+jcol,
		      99, MPI_COMM_WORLD, &status);
           fwrite(row_out,sizeof(float),Mx-2,tfile);
        }
        MPI_Barrier( MPI_COMM_WORLD);

     }
   }
}


void time_step()
{
   int ix, iy, iz, isplit, isub_split;
   float delt_spl, tdx, tdy;
   float dx, dy;
   float gdxdt, gdydt, rgdxdt, rgdydt;
   float dcx,dcy,dcxm,dcym;

   dcxm = Diff_Mks/(Delx * Delx);
   dcym = Diff_Mks/(Dely * Dely);
   dcx = Diff_Ratio*Diff_Mks/(Delx * Delx);
   dcy = Diff_Ratio*Diff_Mks/(Dely * Dely);

 /* Second order conservative dissipation on scalars*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         dts[iz][iy][ix] = 0.0;
         dtt[iz][iy][ix] = 0.0;

         if(h[iz][iy][ix] > 1.0)
         {
            if(h[iz][iy][ix-1] > 1.0)
            {
              dts[iz][iy][ix] -= dcx * (sl[iz][iy][ix] - sl[iz][iy][ix-1]);
              dtt[iz][iy][ix] -= dcx * (tl[iz][iy][ix] - tl[iz][iy][ix-1]);
            }

	        if(h[iz][iy][ix+1] > 1.0)
            {
	          dts[iz][iy][ix] -= dcx * (sl[iz][iy][ix] - sl[iz][iy][ix+1]);
	          dtt[iz][iy][ix] -= dcx * (tl[iz][iy][ix] - tl[iz][iy][ix+1]);
            }

            if(h[iz][iy-1][ix] > 1.0)
            {
	          dts[iz][iy][ix] -= dcy * (sl[iz][iy][ix] - sl[iz][iy-1][ix]);
	          dtt[iz][iy][ix] -= dcy * (tl[iz][iy][ix] - tl[iz][iy-1][ix]);
            }

	        if(h[iz][iy+1][ix] > 1.0)
            {
	          dts[iz][iy][ix] -= dcy * (sl[iz][iy][ix] - sl[iz][iy+1][ix]);
	          dtt[iz][iy][ix] -= dcy * (tl[iz][iy][ix] - tl[iz][iy+1][ix]);
            }
         }
      }
   }
 }

 /* Second order conservative dissipation on momentum*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         dtu[iz][iy][ix] = 0.0;
         dtv[iz][iy][ix] = 0.0;

         if(uvh[iz][iy][ix] > 1.0)
         {
            if(uvh[iz][iy][ix-1] > 1.0)
            {
              dtu[iz][iy][ix] -= dcxm * (ul[iz][iy][ix] - ul[iz][iy][ix-1]);
              dtv[iz][iy][ix] -= dcxm * (vl[iz][iy][ix] - vl[iz][iy][ix-1]);
            }

	        if(uvh[iz][iy][ix+1] > 1.0)
            {
	          dtu[iz][iy][ix] -= dcxm * (ul[iz][iy][ix] - ul[iz][iy][ix+1]);
	          dtv[iz][iy][ix] -= dcxm * (vl[iz][iy][ix] - vl[iz][iy][ix+1]);
            }

            if(uvh[iz][iy-1][ix] > 1.0)
            {
	          dtu[iz][iy][ix] -= dcym * (ul[iz][iy][ix] - ul[iz][iy-1][ix]);
	          dtv[iz][iy][ix] -= dcym * (vl[iz][iy][ix] - vl[iz][iy-1][ix]);
            }

	        if(uvh[iz][iy+1][ix] > 1.0)
            {
	          dtu[iz][iy][ix] -= dcym * (ul[iz][iy][ix] - ul[iz][iy+1][ix]);
	          dtv[iz][iy][ix] -= dcym * (vl[iz][iy][ix] - vl[iz][iy+1][ix]);
            }
         }
      }
   }
 }

/* wind stress forcing on points with full dynamics */

 for (iy = 1; iy < My-1; iy++)
 {
   for (ix = 1; ix < Mx-1; ix++)
   {
	 if(uvh[Nz-1][iy][ix] >1.0)
	 {
	    dtu[Nz-1][iy][ix] += taux[iy][ix];
        dtv[Nz-1][iy][ix] += tauy[iy][ix];
     }
   }
 }

/* Surface forcing on T,S */

 for (iy = 1; iy < My-1; iy++)
 {
   for (ix = 1; ix < Mx-1; ix++)
   {
	 if(h[Nz-1][iy][ix] >1.0)
	 {
	    dtt[Nz-1][iy][ix] += Restore * (annual_temp[Nz-1][iy][ix] - tl[Nz-1][iy][ix]);
        dts[Nz-1][iy][ix] += Restore * (annual_salt[Nz-1][iy][ix] - sl[Nz-1][iy][ix]);
	    t_forcing[iy][ix] += Delt * Restore * (annual_temp[Nz-1][iy][ix] - tl[Nz-1][iy][ix]);
        s_forcing[iy][ix] += Delt * Restore * (annual_salt[Nz-1][iy][ix] - sl[Nz-1][iy][ix]);
     }
   }
 }

/* Calculate Box u-advection dt's u,v,s,t.  Accumulate into dt arrays. */

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      double advc, advc1, advc2;

      advc = geo3[iy]*1.0/(4.0 *Delx);

      for (ix = 1; ix < Mx-1; ix++)
      {
	 if(h[iz][iy][ix] >1.0)
	 {
	    if(h[iz][iy][ix-1] >1.0)
           advc1 = advc * (u[iz][iy][ix-1] + u[iz][iy][ix]);
        else 
           advc1 = 0.0;

	    if(h[iz][iy][ix+1] >1.0)
  	       advc2 = advc * (u[iz][iy][ix] + u[iz][iy][ix+1]);
        else 
           advc2 = 0.0;

	    dtu[iz][iy][ix] += advc1 * (u[iz][iy][ix-1] - u[iz][iy][ix])
	        + advc2 * (u[iz][iy][ix] - u[iz][iy][ix+1]);

	    dtv[iz][iy][ix] += advc1 * (v[iz][iy][ix-1] - v[iz][iy][ix])
	        + advc2 * (v[iz][iy][ix] - v[iz][iy][ix+1]);

	    dts[iz][iy][ix] += advc1 * (s[iz][iy][ix-1] - s[iz][iy][ix])
	        + advc2 * (s[iz][iy][ix] - s[iz][iy][ix+1]);

	    dtt[iz][iy][ix] += advc1 * (t[iz][iy][ix-1] - t[iz][iy][ix])
	        + advc2 * (t[iz][iy][ix] - t[iz][iy][ix+1]);
	 }
      }
   }
 }

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      double advc, advc1, advc2;

      advc = 1.0/(4.0 *Dely);

      for (ix = 1; ix < Mx-1; ix++)
      {
	 if(h[iz][iy][ix] >1.0)
	 {
	    if(h[iz][iy-1][ix] >1.0)
           advc1 = advc * (v[iz][iy-1][ix] + v[iz][iy][ix]);
        else 
           advc1 = 0.0;

	    if(h[iz][iy+1][ix] >1.0)
  	       advc2 = advc * (v[iz][iy][ix] + v[iz][iy+1][ix]);
        else 
           advc2 = 0.0;

	    dtu[iz][iy][ix] += advc1 * (u[iz][iy-1][ix] - u[iz][iy][ix])
	        + advc2 * (u[iz][iy][ix] - u[iz][iy+1][ix]);

	    dtv[iz][iy][ix] += advc1 * (v[iz][iy-1][ix] - v[iz][iy][ix])
	        + advc2 * (v[iz][iy][ix] - v[iz][iy+1][ix]);

	    dts[iz][iy][ix] += advc1 * (s[iz][iy-1][ix] - s[iz][iy][ix])
	        + advc2 * (s[iz][iy][ix] - s[iz][iy+1][ix]);

	    dtt[iz][iy][ix] += advc1 * (t[iz][iy-1][ix] - t[iz][iy][ix])
	        + advc2 * (t[iz][iy][ix] - t[iz][iy+1][ix]);
	 }
      }
   }
 }

   /* setup vx, wx, tx, sx for time split step */

   for (iz = 0; iz < Mz; iz++)
   {
     for (iy = 0; iy < My; iy++)
     {
       for (ix = 0; ix < Mx; ix++)
       {
	      ux[iz][iy][ix] = ul[iz][iy][ix];

	      vx[iz][iy][ix] = vl[iz][iy][ix];

	      tx[iz][iy][ix] = tl[iz][iy][ix];

	      sx[iz][iy][ix] = sl[iz][iy][ix];
       }
     }
   }
   /* setup ptopx for time split step */

     for (iy = 0; iy < My; iy++)
     {
       for (ix = 0; ix < Mx; ix++)
       {
	      ptopx[iy][ix] = ptopl[iy][ix];
       }
     }

   /* caluculate mixing parameters before split */

   for (iz = Mz-1; iz > 0; iz--) /* calculate downward mixing  not iz=0 */
   {
     for (iy = 1; iy < My-1; iy++)
     {
       for (ix = 1; ix < Mx-1; ix++)
       {
	  if(h[iz][iy][ix] > 1.0)
          {  float shrsq,rr2,rr4,rr8;
             if(h[iz-1][iy][ix] > 1.0) /* interior layer */
             {
                nsq[iz][iy][ix] = 9.8*
                  (density(sx[iz-1][iy][ix],tx[iz-1][iy][ix],iz)-
                   density(sx[iz][iy][ix],tx[iz][iy][ix],iz))
                  /(0.5*(h[iz-1][iy][ix]+h[iz][iy][ix]));
             }
             else /* bottom layer */
             {
                if(iz < (Mz-1)) nsq[iz][iy][ix] = nsq[iz+1][iy][ix];
                else nsq[iz][iy][ix] = 3.0e-5; /* also top layer */
             }
             if(nsq[iz][iy][ix] < 1.0e-8) nsq[iz][iy][ix] = 1.0e-8; /* clamp */
             if(uvh[iz][iy][ix] > 1.0) /* cell with active vels */
             {
                if(uvh[iz-1][iy][ix] > 1.0)
                {
                   shrsq = (ux[iz][iy][ix]-ux[iz-1][iy][ix])*
                           (ux[iz][iy][ix]-ux[iz-1][iy][ix])
                          +(vx[iz][iy][ix]-vx[iz-1][iy][ix])*
                           (vx[iz][iy][ix]-vx[iz-1][iy][ix]);
                }
                else
                {
                   shrsq = ux[iz][iy][ix]*ux[iz][iy][ix]
                          +vx[iz][iy][ix]*vx[iz][iy][ix];
                }
                shrsq /= 0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]);
                shrsq /= 0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]);
             }
             else shrsq=0.0;
             if(shrsq < 0.01*nsq[iz][iy][ix]) shrsq = 0.01*nsq[iz][iy][ix]; /* clamp */
             risq[iz][iy][ix] = nsq[iz][iy][ix]/shrsq;
             if(risq[iz][iy][ix] < 0.01) risq[iz][iy][ix] = 0.01;
             rr2 = 1.0/risq[iz][iy][ix];
             rr4 = rr2*rr2;
             rr8 = rr4*rr4;
             kuv[iz][iy][ix] = (7.38e-8 * rr8)+1.0e-4;
             if(kuv[iz][iy][ix] > 0.01) kuv[iz][iy][ix] = 0.01;
             kts[iz][iy][ix] = (2.75e-8 * rr8)+1.0e-5;
             if(kts[iz][iy][ix] > 0.01) kts[iz][iy][ix] = 0.01;
          } 
       }
     }
   }

   iz=0; /* calcualte downward mixing, bottom layer */
   {
     for (iy = 1; iy < My-1; iy++)
     {
       for (ix = 1; ix < Mx-1; ix++)
       {
	  if(h[iz][iy][ix] > 1.0)
          {  float shrsq,rr2,rr4,rr8;

             nsq[iz][iy][ix] = nsq[iz+1][iy][ix];

             if(uvh[iz][iy][ix] > 1.0) /* cell with active vels */
             {
                shrsq = ux[iz][iy][ix]*ux[iz][iy][ix]
                       +vx[iz][iy][ix]*vx[iz][iy][ix];
                shrsq /= 0.5*uvh[iz][iy][ix];
                shrsq /= 0.5*uvh[iz][iy][ix];
             }
             else shrsq=0.0;
             if(shrsq < 0.01*nsq[iz][iy][ix]) shrsq = 0.01*nsq[iz][iy][ix]; /* clamp */
             risq[iz][iy][ix] = nsq[iz][iy][ix]/shrsq;
             if(risq[iz][iy][ix] < 0.01) risq[iz][iy][ix] = 0.01;
             rr2 = 1.0/risq[iz][iy][ix];
             rr4 = rr2*rr2;
             rr8 = rr4*rr4;
             kuv[iz][iy][ix] = (7.38e-8 * rr8)+1.0e-4;
             if(kuv[iz][iy][ix] > 0.01) kuv[iz][iy][ix] = 0.01;
             kts[iz][iy][ix] = (2.75e-8 * rr8)+1.0e-5;
             if(kts[iz][iy][ix] > 0.01) kts[iz][iy][ix] = 0.01;
          } 
       }
     }
   }

   delt_spl = 2.0 * Delt/((float) N_Split);

    for (iz = 0; iz < Mz-1; iz++) /* vertical advection of scalars - gradient calculation */
    { float dz;
      for (iy = 1; iy < My-1; iy++)
      {
	 for (ix = 1; ix < Mx-1; ix++)
	 {
        if((h[iz][iy][ix] >1.0) && (h[iz+1][iy][ix] >1.0))
        {
           dz = 1/(0.5*(h[iz][iy][ix]+h[iz+1][iy][ix]));

	       dzt[iz][iy][ix] = dz*(t[iz+1][iy][ix]-t[iz][iy][ix]);
	       dzs[iz][iy][ix] = dz*(s[iz+1][iy][ix]-s[iz][iy][ix]);
        }
        if((uvh[iz][iy][ix] >1.0) && (uvh[iz+1][iy][ix] >1.0))
        {
           dz = 1/(0.5*(h[iz][iy][ix]+h[iz+1][iy][ix]));

	       dzu[iz][iy][ix] = dz*(u[iz+1][iy][ix]-u[iz][iy][ix]);
	       dzv[iz][iy][ix] = dz*(v[iz+1][iy][ix]-v[iz][iy][ix]);
        }
	 }
      }
    }

   for (isplit = 0; isplit < N_Split; isplit++) /* time split hydrostatic integration */
   { float dxdt,dydt;

    for (iz = 0; iz < Mz; iz++) /* calculate vertical velocity */
    {  float cwx, cwy;

      cwy = 1.0/(2.0*Dely);
 
      for (iy = 1; iy < My-1; iy++)
      {
      cwx = geo3[iy]*1.0/(2.0*Delx);
	 for (ix = 1; ix < Mx-1; ix++)
	 {
        if(h[iz][iy][ix] >1.0)
        {
	      if(iz == 0)
          {
                w[iz][iy][ix] =
                    -cwx*(uvh[iz][iy][ix+1]*ux[iz][iy][ix+1] 
                        - uvh[iz][iy][ix-1]*ux[iz][iy][ix-1])
	                -cwy*(geo2[iy]*uvh[iz][iy+1][ix]*vx[iz][iy+1][ix]
                        - geo1[iy]*uvh[iz][iy-1][ix]*vx[iz][iy-1][ix]);
           } 
	       else
           {
                w[iz][iy][ix] = w[iz-1][iy][ix]
                     -cwx*(uvh[iz][iy][ix+1]*ux[iz][iy][ix+1] 
                         - uvh[iz][iy][ix-1]*ux[iz][iy][ix-1])
	                 -cwy*(geo2[iy]*uvh[iz][iy+1][ix]*vx[iz][iy+1][ix]
                         - geo1[iy]*uvh[iz][iy-1][ix]*vx[iz][iy-1][ix]);
           }
        }
        else w[iz][iy][ix] = 0.0;
	 }
      }
    }

    for (iz = 0; iz < Mz-1; iz++) /* vertical advection of scalars - update */
    { float dt;
      for (iy = 1; iy < My-1; iy++)
      {
	 for (ix = 1; ix < Mx-1; ix++)
	 {
        if((h[iz][iy][ix] >1.0) && (h[iz+1][iy][ix] >1.0))
        {
           dt = 0.5*delt_spl;

	       tx[iz][iy][ix] -= dt*w[iz][iy][ix]*dzt[iz][iy][ix];
	       tx[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzt[iz][iy][ix];

	       sx[iz][iy][ix] -= dt*w[iz][iy][ix]*dzs[iz][iy][ix];
	       sx[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzs[iz][iy][ix];
        }
        if((uvh[iz][iy][ix] >1.0) && (uvh[iz+1][iy][ix] >1.0))
        {
           dt = 0.5*delt_spl;

	       ux[iz][iy][ix] -= dt*w[iz][iy][ix]*dzu[iz][iy][ix];
	       ux[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzu[iz][iy][ix];

	       vx[iz][iy][ix] -= dt*w[iz][iy][ix]*dzv[iz][iy][ix];
	       vx[iz+1][iy][ix] -= dt*w[iz][iy][ix]*dzv[iz][iy][ix];
        }
	 }
      }
    }

    for (iz = 0; iz < Mz; iz++) /* time tendancies for scalars */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	   for (ix = 1; ix < Mx-1; ix++)
	   {
         if(h[iz][iy][ix] >1.0)
         {
	       tx[iz][iy][ix] += delt_spl * dtt[iz][iy][ix];
	       sx[iz][iy][ix] += delt_spl * dts[iz][iy][ix];
         }
	   }
      }
    }

    for (iy = 1; iy < My-1; iy++)
    {
	  for (ix = 1; ix < Mx-1; ix++)
	  {
         p[Nz-1][iy][ix] = 0.0;
	  }
    }

    for (iz = Nz-2; iz >= 0; iz--) /* hydrostatic pressure integration */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	    for (ix = 1; ix < Mx-1; ix++)
	    {
	      p[iz][iy][ix] = p[iz+1][iy][ix]+
             9.8*0.5*(
               h[iz][iy][ix]*density(sx[iz][iy][ix],tx[iz][iy][ix],iz)
             + h[iz+1][iy][ix]*density(sx[iz+1][iy][ix],tx[iz+1][iy][ix],iz+1));
	    }
      }
    }

    wrap_qz(p);

    for (iy = 0; iy < My; iy++) /* split barotropic mode u,v */
    {
      for (ix = 0; ix < Mx; ix++)
      {
        ubaro[iy][ix] = 0.0;
        vbaro[iy][ix] = 0.0;
	    for (iz = 0; iz < Mz; iz++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
	          ubaro[iy][ix] += h[iz][iy][ix]*ux[iz][iy][ix];
	          vbaro[iy][ix] += h[iz][iy][ix]*vx[iz][iy][ix];
           }
	    }
        ubaro[iy][ix] = ruvdepth[iy][ix] * ubaro[iy][ix];
        vbaro[iy][ix] = ruvdepth[iy][ix] * vbaro[iy][ix];
	    for (iz = 0; iz < Mz; iz++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
              ux[iz][iy][ix] -= ubaro[iy][ix];
              vx[iz][iy][ix] -= vbaro[iy][ix];
           }
	    }
      }
    }

    dy = 1.0/(2.0*Dely);

    for (iy = 1; iy < My-1; iy++) /* split barotropic mode dtu, dtv including pgf */
    { float pgfx, pgfy;
      dx = geo3[iy]*1.0/(2.0*Delx);
      for (ix = 1; ix < Mx-1; ix++)
      {
        dtubaro[iy][ix] = 0.0;
        dtvbaro[iy][ix] = 0.0;
	    for (iz = 0; iz < Mz; iz++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
	          pgfx = -dx * (p[iz][iy][ix+1] - p[iz][iy][ix-1]);
	          dtubaro[iy][ix] += h[iz][iy][ix] * (pgfx + dtu[iz][iy][ix]);

	          pgfy = -dy * (p[iz][iy+1][ix] - p[iz][iy-1][ix]);
	          dtvbaro[iy][ix] += h[iz][iy][ix] * (pgfy + dtv[iz][iy][ix]);
           }
	    }
        dtubaro[iy][ix] = ruvdepth[iy][ix] * dtubaro[iy][ix];
        dtvbaro[iy][ix] = ruvdepth[iy][ix] * dtvbaro[iy][ix];
      }
    }


    dydt = delt_spl/(2.0*Dely);

    for (iz = 0; iz < Mz; iz++)
    {
      for (iy = 1; iy < My-1; iy++)
      { float dtf, a, b, r, pgfx, pgfy;
 
     dxdt = geo3[iy]*delt_spl/(2.0*Delx);

	 dtf = delt_spl*f[iy];
     r = 1.0/(1.0+dtf*dtf); /* norm for semi-implicit coriolis */

	 for (ix = 1; ix < Mx-1; ix++)
	 {
	    if(uvh[iz][iy][ix] >1.0) /* no velocities in dry boxes */
	    {
	         pgfx = -dxdt * (p[iz][iy][ix+1] - p[iz][iy][ix-1]);
             a = delt_spl * (dtu[iz][iy][ix] - dtubaro[iy][ix]);

	         pgfy = -dydt * (p[iz][iy+1][ix] - p[iz][iy-1][ix]);
             b = delt_spl * (dtv[iz][iy][ix] - dtvbaro[iy][ix]);

	         ux[iz][iy][ix] = r*(ux[iz][iy][ix] + pgfx + 
                                 dtf*(vx[iz][iy][ix]+ b + pgfy) + a);

 	         vx[iz][iy][ix] += pgfy - dtf*ux[iz][iy][ix] + b;
	    }
	 }
      }
    }

 for (isub_split = 0; isub_split < Sub_Split; isub_split++) /* barotropic sub-cycle */
 { float delt_sub;

    delt_sub = delt_spl/Sub_Split;

   for (iy = 1; iy < My-1; iy++) /* Barotropic Pressure field predict*/
   { float cwx, cwy;

     cwx = geo3[iy]*1.0/(2.0*Delx);
     cwy = 1.0/(2.0*Dely);

	 for (ix = 1; ix < Mx-1; ix++)
	 {
	    ptstar[iy][ix] = ptopx[iy][ix]-G*delt_sub*
                           (cwx*(uvdepth[iy][ix+1]*ubaro[iy][ix+1] 
                               - uvdepth[iy][ix-1]*ubaro[iy][ix-1])
	                       +cwy*(geo2[iy]*uvdepth[iy+1][ix]*vbaro[iy+1][ix]
                               - geo1[iy]*uvdepth[iy-1][ix]*vbaro[iy-1][ix]));
	 }
   }

   dydt = delt_sub/(2.0*Dely);

   for (iy = 1; iy < My-1; iy++) /* barotropic u,v predict*/
   { float dtf, a, b, r, pgfx, pgfy;

     dxdt = geo3[iy]*delt_sub/(2.0*Delx);

	 dtf = delt_sub*f[iy];
     r = 1.0/(1.0 + dtf*dtf); /* norm for semi-implicit coriolis */

	 for (ix = 1; ix < Mx-1; ix++)
	 {

	    if(uvdepth[iy][ix] > 1.0)
        {
	         pgfx = -dxdt * (ptopx[iy][ix+1] - ptopx[iy][ix-1]);
             a = delt_sub * dtubaro[iy][ix];

	         pgfy = -dydt * (ptopx[iy+1][ix] - ptopx[iy-1][ix]);
             b = delt_sub * dtvbaro[iy][ix];

	         ubstar[iy][ix] = ubaro[iy][ix] + pgfx + dtf*vbaro[iy][ix] + a;

 	         vbstar[iy][ix] = vbaro[iy][ix] + pgfy - dtf*ubaro[iy][ix] + b;
         }
	  }
   }

   wrap_q3(ubstar, vbstar, ptstar);

   for (iy = 1; iy < My-1; iy++) /* Barotropic Pressure field correct*/
   { float cwx, cwy;

     cwx = geo3[iy]*1.0/(2.0*Delx);
     cwy = 1.0/(2.0*Dely);

	 for (ix = 1; ix < Mx-1; ix++)
	 {
	    ptopx[iy][ix] = ptopx[iy][ix]-G*delt_sub*
                           (cwx*(uvdepth[iy][ix+1]*ubstar[iy][ix+1] 
                               - uvdepth[iy][ix-1]*ubstar[iy][ix-1])
	                       +cwy*(geo2[iy]*uvdepth[iy+1][ix]*vbstar[iy+1][ix]
                               - geo1[iy]*uvdepth[iy-1][ix]*vbstar[iy-1][ix]));
	 }
   }

   dydt = delt_sub/(2.0*Dely);

   for (iy = 1; iy < My-1; iy++) /* barotropic u,v correct */
   { float dtf, a, b, r, pgfx, pgfy;

     dxdt = geo3[iy]*delt_sub/(2.0*Delx);

	 dtf = delt_sub*f[iy];

	 for (ix = 1; ix < Mx-1; ix++)
	 {

	    if(uvdepth[iy][ix] > 1.0)
        {
	         pgfx = -dxdt * (ptstar[iy][ix+1] - ptstar[iy][ix-1]);
             a = delt_sub * dtubaro[iy][ix];

	         pgfy = -dydt * (ptstar[iy+1][ix] - ptstar[iy-1][ix]);
             b = delt_sub * dtvbaro[iy][ix];

	         ubaro[iy][ix] = ubaro[iy][ix] + pgfx + dtf*vbstar[iy][ix] + a;

 	         vbaro[iy][ix] = vbaro[iy][ix] + pgfy - dtf*ubstar[iy][ix] + b;
         }
	  }
   }

  wrap_q3(ubaro, vbaro, ptopx);
  filt4p(ptopx);

 } /* end barotropic sub cycle */

    for (iz = 0; iz < Mz; iz++) /* unsplit barotropic mode u,v */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	    for (ix = 1; ix < Mx-1; ix++)
	    {
	       if(uvh[iz][iy][ix] > 1.0)
           {
              ux[iz][iy][ix] += ubaro[iy][ix];
              vx[iz][iy][ix] += vbaro[iy][ix];
           }
	    }
      }
    }

 /* Vertical mixing */

 for (iz = 1; iz < Mz-1; iz++) /* Interior levels */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { float u_up_str,u_down_str,v_up_str,v_down_str,mix;

            if(uvh[iz-1][iy][ix] > 1.0)
            {
              u_down_str = (ux[iz][iy][ix]-ux[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
              v_down_str = (vx[iz][iy][ix]-vx[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
            }
            else
            {
              u_down_str = ux[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
              v_down_str = vx[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
            }
            u_up_str = (ux[iz+1][iy][ix]-ux[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));
            v_up_str = (vx[iz+1][iy][ix]-vx[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));

            u_down_str *= kuv[iz][iy][ix];
            v_down_str *= kuv[iz][iy][ix];
            u_up_str *= kuv[iz+1][iy][ix];
            v_up_str *= kuv[iz+1][iy][ix];

            mix = delt_spl/uvh[iz][iy][ix];
            mix_u[iz][iy][ix] = mix*(u_up_str-u_down_str);
            mix_v[iz][iy][ix] = mix*(v_up_str-v_down_str);
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { float t_up_str,t_down_str,s_up_str,s_down_str,scmix;

            if(h[iz-1][iy][ix] > 1.0)
            {
              t_down_str = (tx[iz][iy][ix]-tx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
              s_down_str = (sx[iz][iy][ix]-sx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
            }
            else
            {
              t_down_str = 0.0;
              s_down_str = 0.0;
            }
            t_up_str = (tx[iz+1][iy][ix]-tx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));
            s_up_str = (sx[iz+1][iy][ix]-sx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));

            t_down_str *= kts[iz][iy][ix];
            s_down_str *= kts[iz][iy][ix];
            t_up_str *= kts[iz+1][iy][ix];
            s_up_str *= kts[iz+1][iy][ix];

            scmix = delt_spl/h[iz][iy][ix];
            mix_t[iz][iy][ix] = scmix*(t_up_str-t_down_str);
            mix_s[iz][iy][ix] = scmix*(s_up_str-s_down_str);
         }
      }
   }
 }

 iz = 0; /* Bottom level */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { float u_up_str,u_down_str,v_up_str,v_down_str,mix;

            u_down_str = ux[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
            v_down_str = vx[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));

            u_up_str = (ux[iz+1][iy][ix]-ux[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));
            v_up_str = (vx[iz+1][iy][ix]-vx[iz][iy][ix])/
                       (0.5*(uvh[iz+1][iy][ix]+uvh[iz][iy][ix]));

            u_down_str *= kuv[iz][iy][ix];
            v_down_str *= kuv[iz][iy][ix];
            u_up_str *= kuv[iz+1][iy][ix];
            v_up_str *= kuv[iz+1][iy][ix];

            mix = delt_spl/uvh[iz][iy][ix];
            mix_u[iz][iy][ix] = mix*(u_up_str-u_down_str);
            mix_v[iz][iy][ix] = mix*(v_up_str-v_down_str);
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { float t_up_str,s_up_str,scmix;

            t_up_str = (tx[iz+1][iy][ix]-tx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));
            s_up_str = (sx[iz+1][iy][ix]-sx[iz][iy][ix])/
                       (0.5*(h[iz+1][iy][ix]+h[iz][iy][ix]));

            t_up_str *= kts[iz+1][iy][ix];
            s_up_str *= kts[iz+1][iy][ix];

            scmix = delt_spl/h[iz][iy][ix];
            mix_t[iz][iy][ix] = scmix*(t_up_str);
            mix_s[iz][iy][ix] = scmix*(s_up_str);
         }
      }
   }
 }

 iz=Nz-1; /* Top level */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { float u_down_str,v_down_str,mix;

            if(uvh[iz-1][iy][ix] > 1.0)
            {
              u_down_str = (ux[iz][iy][ix]-ux[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
              v_down_str = (vx[iz][iy][ix]-vx[iz-1][iy][ix])/
                           (0.5*(uvh[iz][iy][ix]+uvh[iz-1][iy][ix]));
            }
            else
            {
              u_down_str = ux[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
              v_down_str = vx[iz][iy][ix]/(0.5*(uvh[iz][iy][ix]));
            }

            u_down_str *= kuv[iz][iy][ix];
            v_down_str *= kuv[iz][iy][ix];

            mix = delt_spl/uvh[iz][iy][ix];
            mix_u[iz][iy][ix] = mix*(-u_down_str);
            mix_v[iz][iy][ix] = mix*(-v_down_str);
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { float t_down_str,s_down_str,scmix;

            if(h[iz-1][iy][ix] > 1.0)
            {
              t_down_str = (tx[iz][iy][ix]-tx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
              s_down_str = (sx[iz][iy][ix]-sx[iz-1][iy][ix])/
                           (0.5*(h[iz][iy][ix]+h[iz-1][iy][ix]));
            }
            else
            {
              t_down_str = 0.0;
              s_down_str = 0.0;
            }

            t_down_str *= kts[iz][iy][ix];
            s_down_str *= kts[iz][iy][ix];

            scmix = delt_spl/h[iz][iy][ix];
            mix_t[iz][iy][ix] = scmix*(-t_down_str);
            mix_s[iz][iy][ix] = scmix*(-s_down_str);
         }
      }
   }
 }

 for (iz = 0; iz < Mz; iz++) /* mixing update */
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         { 
            ux[iz][iy][ix] += mix_u[iz][iy][ix];
            vx[iz][iy][ix] += mix_v[iz][iy][ix];
         }
/* mix scalars */
         if(h[iz][iy][ix] > 1.0)
         { 
            tx[iz][iy][ix] += mix_t[iz][iy][ix];
            sx[iz][iy][ix] += mix_s[iz][iy][ix];
         }
      }
   }
 }
    wrap_qz(ux);
    wrap_qz(vx);
    wrap_qz(tx);
    wrap_qz(sx);
   } /* end timesplit part */

    for (iz = 0; iz < Mz; iz++) /* calculate full pressure field for diagnostics */
    {
      for (iy = 1; iy < My-1; iy++)
      {
	    for (ix = 1; ix < Mx-1; ix++)
	    {
	      p[iz][iy][ix] += ptopx[iy][ix];
	    }
      }
    }

    for (iy = 0; iy < My; iy++) /* convective adjustment */
    {
      for (ix = 0; ix < Mx; ix++)
      { int mixing, pass_count;

        mixing = 1;
        pass_count = 0;
        while(mixing && (pass_count++ < Nz))
        {
          mixing = 0;
	      for (iz = Mz-2; iz >= 0; iz--)
	      { float d1 ,d2 , snew, tnew;
            if((h[iz][iy][ix] > 1.0) && (h[iz+1][iy][ix]) > 1.0)
            {
               d1 = density(sx[iz][iy][ix],tx[iz][iy][ix],iz);
               d2 = density(sx[iz+1][iy][ix],tx[iz+1][iy][ix],iz);
               if(d2 > (d1 + 1.0e-5))
               {
                  mixing = 1;

                  snew = (sx[iz][iy][ix]*h[iz][iy][ix] +
                          sx[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);
                  sx[iz][iy][ix] = sx[iz+1][iy][ix] = snew;

                  tnew = (tx[iz][iy][ix]*h[iz][iy][ix] +
                          tx[iz+1][iy][ix]*h[iz+1][iy][ix])/
                         (h[iz][iy][ix] + h[iz+1][iy][ix]);
                  tx[iz][iy][ix] = tx[iz+1][iy][ix] = tnew;
               }
            }
	      }
	    }
      }
    }

/* Fourth order space filter */

  filt4m(ux);
  filt4m(vx);
  filt4s(tx);
  filt4s(sx);
   
/* Time filter */

   for (iz =  0; iz < Mz; iz++)
   {
     for (iy = 0; iy < My; iy++)
     {
      for (ix = 0; ix < Mx; ix++)
      {
	 ul[iz][iy][ix] = 0.65 * u[iz][iy][ix]
	     + 0.175 * (ul[iz][iy][ix] + ux[iz][iy][ix]);
	 u[iz][iy][ix] = ux[iz][iy][ix];

	 vl[iz][iy][ix] = 0.65 * v[iz][iy][ix]
	     + 0.175 * (vl[iz][iy][ix] + vx[iz][iy][ix]);
	 v[iz][iy][ix] = vx[iz][iy][ix];

	 tl[iz][iy][ix] = 0.65 * t[iz][iy][ix]
	     + 0.175 * (tl[iz][iy][ix] + tx[iz][iy][ix]);
	 t[iz][iy][ix] = tx[iz][iy][ix];

	 sl[iz][iy][ix] = 0.65 * s[iz][iy][ix]
	     + 0.175 * (sl[iz][iy][ix] + sx[iz][iy][ix]);
	 s[iz][iy][ix] = sx[iz][iy][ix];
      }
     }
   }

   for (iy = 0; iy < My; iy++)
   {
     for (ix = 0; ix < Mx; ix++)
     {
	    ptopl[iy][ix] = 0.65 * ptop[iy][ix]
	      + 0.175 * (ptopl[iy][ix] + ptopx[iy][ix]);
	    ptop[iy][ix] = ptopx[iy][ix];
      }
    }

}

float density(float s, float t, int iz)
{
   float den;

/*
 * simple approximate density for salt water.  To be replaced with
 * std polynomical formula when convection details are important.
*/
   den = 0.2*(s-base_s[iz]) - 0.2*(t-base_t[iz]);
   return den/1030.0;
}

void filt4s(float q[Mz][My][Mx])
{
  int ix,iy,iz;
  float damp,work[Mz][My][Mx];

  damp = 0.5/32.0;

 /* Fourth order filter on scalars - pass 1*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         work[iz][iy][ix] = 0.0;

         if(h[iz][iy][ix] > 1.0)
         {
            if(h[iz][iy][ix-1] > 1.0)
            {
              work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix-1]);
            }

	        if(h[iz][iy][ix+1] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix+1]);
            }

            if(h[iz][iy-1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy-1][ix]);
            }

	        if(h[iz][iy+1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy+1][ix]);
            }
         }
      }
   }
 }

 wrap_qz(work);

 /* Fourth order conservative dissipation on scalars - pass2 */

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(h[iz][iy][ix] > 1.0)
         {
            if(h[iz][iy][ix-1] > 1.0)
            {
              q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix-1]);
            }

	        if(h[iz][iy][ix+1] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix+1]);
            }

            if(h[iz][iy-1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy-1][ix]);
            }

	        if(h[iz][iy+1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy+1][ix]);
            }
         }
      }
   }
 }
wrap_qz(q);
}

void filt4m(float q[Mz][My][Mx])
{
  int ix,iy,iz;
  float damp,work[Mz][My][Mx];

  damp = 0.5/32.0;

 /* Fourth order filter on momentum - pass 1*/

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         work[iz][iy][ix] = 0.0;

         if(uvh[iz][iy][ix] > 1.0)
         {
            if(uvh[iz][iy][ix-1] > 1.0)
            {
              work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix-1]);
            }

	        if(uvh[iz][iy][ix+1] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy][ix+1]);
            }

            if(uvh[iz][iy-1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy-1][ix]);
            }

	        if(uvh[iz][iy+1][ix] > 1.0)
            {
	          work[iz][iy][ix] -= (q[iz][iy][ix] - q[iz][iy+1][ix]);
            }
         }
      }
   }
 }

 wrap_qz(work);

 /* Fourth order conservative dissipation on momentum - pass2 */

 for (iz = 0; iz < Mz; iz++)
 {
   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(uvh[iz][iy][ix] > 1.0)
         {
            if(uvh[iz][iy][ix-1] > 1.0)
            {
              q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix-1]);
            }

	        if(uvh[iz][iy][ix+1] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy][ix+1]);
            }

            if(uvh[iz][iy-1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy-1][ix]);
            }

	        if(uvh[iz][iy+1][ix] > 1.0)
            {
	          q[iz][iy][ix] += damp * (work[iz][iy][ix] - work[iz][iy+1][ix]);
            }
         }
      }
   }
 }
wrap_qz(q);
}

void filt4p(float q[My][Mx])
{
  int ix,iy;
  float damp,work[My][Mx];

  damp = 0.5/32.0;

 /* Fourth order filter on pressure - pass 1*/

   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         work[iy][ix] = 0.0;

         if(depth[iy][ix] > 1.0)
         {
            if(depth[iy][ix-1] > 1.0)
            {
              work[iy][ix] -= (q[iy][ix] - q[iy][ix-1]);
            }

	        if(depth[iy][ix+1] > 1.0)
            {
	          work[iy][ix] -= (q[iy][ix] - q[iy][ix+1]);
            }

            if(depth[iy-1][ix] > 1.0)
            {
	          work[iy][ix] -= (q[iy][ix] - q[iy-1][ix]);
            }

	        if(depth[iy+1][ix] > 1.0)
            {
	          work[iy][ix] -= (q[iy][ix] - q[iy+1][ix]);
            }
         }
      }
   }

 wrap_q(work);

 /* Fourth order conservative dissipation on pressure - pass2 */

   for (iy = 1; iy < My-1; iy++)
   {
      for (ix = 1; ix < Mx-1; ix++) 
      {
         if(depth[iy][ix] > 1.0)
         {
            if(depth[iy][ix-1] > 1.0)
            {
              q[iy][ix] += damp * (work[iy][ix] - work[iy][ix-1]);
            }

	        if(depth[iy][ix+1] > 1.0)
            {
	          q[iy][ix] += damp * (work[iy][ix] - work[iy][ix+1]);
            }

            if(depth[iy-1][ix] > 1.0)
            {
	          q[iy][ix] += damp * (work[iy][ix] - work[iy-1][ix]);
            }

	        if(depth[iy+1][ix] > 1.0)
            {
	          q[iy][ix] += damp * (work[iy][ix] - work[iy+1][ix]);
            }
         }
      }
   }
wrap_q(q);
}


/* Begin sp2 communications routines. */

float in_xz[Mz][Mx], out_xz[Mz][Mx];
float in_yz[Mz][My], out_yz[Mz][My];

void wrap_qz(float a[Mz][My][Mx])
{
   int ix, iy, iz;
   int tag=0;

/* rec from south, send north */
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       out_xz[iz][ix] = a[iz][My-2][ix];
     }
   }
   send_and_receive(south, tag, in_xz, Mx*Mz*sizeof(float), north, out_xz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       a[iz][0][ix] = in_xz[iz][ix];
     }
   }

/* rec from north, send south */
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       out_xz[iz][ix] = a[iz][1][ix];
     }
   }
   send_and_receive(north, tag, in_xz, Mx*Mz*sizeof(float), south, out_xz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(ix = 0; ix < Mx; ix++)
     {
       a[iz][My-1][ix] = in_xz[iz][ix];
     }
   }

/* rec from west, send east */
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       out_yz[iz][iy] = a[iz][iy][Mx-2];
     }
   }
   send_and_receive(west, tag, in_yz, My*Mz*sizeof(float), east, out_yz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       a[iz][iy][0] = in_yz[iz][iy];
     }
   }

/* rec from east, send west */
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       out_yz[iz][iy] = a[iz][iy][1];
     }
   }
   send_and_receive(east, tag, in_yz, My*Mz*sizeof(float), west, out_yz);
   for(iz = 0; iz < Mz; iz++)
   {
     for(iy = 0; iy < My; iy++)
     {
       a[iz][iy][Mx-1] = in_yz[iz][iy];
     }
   }
}

float in_3x[3][Mx], out_3x[3][Mx];
float in_3y[3][My], out_3y[3][My];

void wrap_q3(float a[My][Mx], float b[My][Mx], float c[My][Mx])
{
   int ix, iy;
   int tag=0;

/* rec from south, send north */
   for(ix = 0; ix < Mx; ix++)
   {
     out_3x[0][ix] = a[My-2][ix];
     out_3x[1][ix] = b[My-2][ix];
     out_3x[2][ix] = c[My-2][ix];
   }
   send_and_receive(south, tag, in_3x, Mx*3*sizeof(float), north, out_3x);
   for(ix = 0; ix < Mx; ix++)
   {
     a[0][ix] = in_3x[0][ix];
     b[0][ix] = in_3x[1][ix];
     c[0][ix] = in_3x[2][ix];
   }

/* rec from north, send south */
   for(ix = 0; ix < Mx; ix++)
   {
     out_3x[0][ix] = a[1][ix];
     out_3x[1][ix] = b[1][ix];
     out_3x[2][ix] = c[1][ix];
   }
   send_and_receive(north, tag, in_3x, Mx*3*sizeof(float), south, out_3x);
   for(ix = 0; ix < Mx; ix++)
   {
     a[My-1][ix] = in_3x[0][ix];
     b[My-1][ix] = in_3x[1][ix];
     c[My-1][ix] = in_3x[2][ix];
   }

/* rec from west, send east */
   for(iy = 0; iy < My; iy++)
   {
     out_3y[0][iy] = a[iy][Mx-2];
     out_3y[1][iy] = b[iy][Mx-2];
     out_3y[2][iy] = c[iy][Mx-2];
   }
   send_and_receive(west, tag, in_3y, My*3*sizeof(float), east, out_3y);
   for(iy = 0; iy < My; iy++)
   {
     a[iy][0] = in_3y[0][iy];
     b[iy][0] = in_3y[1][iy];
     c[iy][0] = in_3y[2][iy];
   }

/* rec from east, send west */
   for(iy = 0; iy < My; iy++)
   {
     out_3y[0][iy] = a[iy][1];
     out_3y[1][iy] = b[iy][1];
     out_3y[2][iy] = c[iy][1];
   }
   send_and_receive(east, tag, in_3y, My*3*sizeof(float), west, out_3y);
   for(iy = 0; iy < My; iy++)
   {
     a[iy][Mx-1] = in_3y[0][iy];
     b[iy][Mx-1] = in_3y[1][iy];
     c[iy][Mx-1] = in_3y[2][iy];
   }
}

void wrap_q(a)
float a[My][Mx];
{
   int src, dest, msg_len, tag=0, ellen, elcnt, stride;
   char *msg_in, *msg_out;

   msg_len = Mx * sizeof(float);

   msg_in = (void *) &(a[0][0]); /* rec from south, send north */
   msg_out = (void *) &(a[My-2][0]);

   send_and_receive(south, tag, msg_in, msg_len, north, msg_out);

   msg_in = (void *) &(a[My-1][0]); /* rec from north, send south */
   msg_out = (void *) &(a[1][0]);

   send_and_receive(north, tag, msg_in, msg_len, south, msg_out);

   msg_in = (void *) &(a[0][0]);  /* rec from west, send east */
   msg_out = (void *) &(a[0][Mx-2]);
/* 
 * defined in mpi_edge_type
 *
 * ellen = sizeof(float);
 * elcnt = My;
 * stride = Mx * ellen;
*/

   send_and_receive_v(west, tag, msg_in, ellen, stride, elcnt, east, msg_out);

   msg_in = (void *) &(a[0][Mx-1]);
   msg_out = (void *) &(a[0][1]);
/* 
 * defined in mpi_edge_type
 *
 * ellen = sizeof(float);
 * elcnt = My;
 * stride = Mx * ellen;
*/

   send_and_receive_v(east, tag, msg_in, ellen, stride, elcnt, west, msg_out);
}

send_and_receive(
      int source_node, int tag, void *inbuffer, int buf_len,
      int dest_node, void *outbuffer)
{
    MPI_Status status;
    int result;

#ifdef DEBUG
    if(source_node != dest_node) {
	printf("SEND_RECEIVE: Source and dest node differ: %d %d\n",
	       source_node, dest_node);
	exit(1);
    }
#endif

    result = MPI_Sendrecv(outbuffer, buf_len, MPI_CHAR, dest_node, tag,
			  inbuffer, buf_len, MPI_CHAR, source_node, tag,
			  MPI_COMM_WORLD, &status);

    if(result != MPI_SUCCESS) {
	printf("SEND_RECEIVE: Return code %d\n",result);
	exit(1);
    }
}

send_and_receive_v(
      int source_node, int tag, void *inbuffer,
      int elem_len, int stride, int elem_count,
      int dest_node, void *outbuffer)
{
    MPI_Status status;
    int result;

#ifdef DEBUG
    if(source_node != dest_node) {
	printf("SEND_RECEIVE_V: Source and dest node differ: %d %d\n",
	       source_node, dest_node);
	exit(1);
    }
#endif

    result = MPI_Sendrecv(outbuffer, 1, mpi_edge_type, dest_node, tag,
			  inbuffer, 1, mpi_edge_type, source_node, tag,
			  MPI_COMM_WORLD, &status);

    if(result != MPI_SUCCESS) {
	printf("SEND_RECEIVE_V: Return code %d\n",result);
	exit(1);
    }
}

/* End sp2 communications routines */




int global_open(char *file, int flags, int mode)
{
    int fd;

    if(my_address == 0) {
	if((fd = open(file, flags, mode)) < 0) {
	    printf("File open error: %s\n",file);
	    exit(1);
	}
	printf("File open: %s\n",file);
	return(fd);
    }
    else
	return(-1);
}
