/************* dslash_lean_forwards.c *******************************/
/* MIMD version 6 */
/* Version to use less memory - do things one direction at a time */

/* Version for "valence approximation":
   -- in dslash (isign=PLUS) cuts use of fermions at x+TUP
   -- in dslash^dagger (isign=MINUS) cuts use of fermions at x-TUP
*/

/*
	dslash(F_OFFSET(psi),F_OFFSET(mp),isign,l_parity);
Compute SUM_dirs ( 
    ( 1 + isign*gamma[dir] ) * U(x,dir) * src(x+dir)
  + ( 1 - isign*gamma[dir] ) * U_adj(x-dir,dir) * src(x-dir)
)

*/

#include "generic_wilson_includes.h"
#include "../include/loopend.h"

#ifdef T3E
void prefetch_vector( su3_vector * );
void prefetch_wvec( wilson_vector * );
void prefetch_hwvec( half_wilson_vector * );
void prefetch_matrix( su3_matrix * );
#endif

void dslash(field_offset src,field_offset dest,int isign,int parity)
{
half_wilson_vector hwv;

register int i,end;
register site *s;
register int dir,otherparity;
msg_tag *tag[2];

    switch(parity) {
	case EVEN:      otherparity=ODD; break;
	case ODD:       otherparity=EVEN; break;
	case EVENANDODD:        otherparity=EVENANDODD; break;
    }

    for(dir=XUP;dir<=TUP;dir++){
        /* Take Wilson projection for src displaced in up direction, gather
           it to "our site" */
        FORSOMEPARITY(i,s,otherparity){
#ifdef T3E
prefetch_wvec( (wilson_vector *)(F_PT((s+1),src)) );
#endif
            wp_shrink( (wilson_vector *)F_PT(s,src), 
		       &(s->htmp[0]), dir, isign);
        } END_LOOP
	tag[0]=start_gather( F_OFFSET(htmp[0]), sizeof(half_wilson_vector),
	    dir, parity, gen_pt[0] );

        /* Take Wilson projection for src displaced in down direction,
        multiply it by adjoint link matrix, gather it "up" */
        FORSOMEPARITY(i,s,otherparity){
#ifdef T3E
prefetch_wvec( (wilson_vector *)(F_PT((s+1),src)) );
prefetch_matrix( &(s->link[dir]) );
#endif
            wp_shrink( (wilson_vector *)F_PT(s,src), 
		      &hwv, dir, -isign);
	    mult_adj_su3_mat_hwvec( &(s->link[dir]), &hwv, &(s->htmp[1]));
        } END_LOOP

	tag[1]=start_gather(F_OFFSET(htmp[1]), 
		sizeof(half_wilson_vector), OPP_DIR(dir),
		parity, gen_pt[1] );


        /* Take Wilson projection for src displaced in up direction, gathered,
		multiply it by link matrix, expand it, and add.
		to dest */
	wait_gather(tag[0]);
#ifndef T3E
        if(dir==XUP){ FORSOMEPARITY(i,s,parity){
	    mult_su3_mat_hwvec( &(s->link[dir]), 
		(half_wilson_vector * )(gen_pt[0][i]), &hwv ); 
	    wp_grow( &hwv, (wilson_vector *)F_PT(s,dest), dir, isign);
        }} END_LOOP
	/* Don't do next for dir=TUP and isign=PLUS, i.e. in dslash! */
        else if(dir<TUP || isign==MINUS){ FORSOMEPARITY(i,s,parity){
	    mult_su3_mat_hwvec( &(s->link[dir]), 
		(half_wilson_vector * )(gen_pt[0][i]), &hwv ); 
	    wp_grow_add( &hwv, (wilson_vector *)F_PT(s,dest), dir, isign);
        }} END_LOOP
#else
    switch(parity){
	case EVEN:
	    i=0; end = even_sites_on_node; break;
	case ODD:
	    i = even_sites_on_node; end = sites_on_node; break;
	case EVENANDODD:
	    i=0; end = sites_on_node; break;
    }
    end--;
    if(dir==XUP){
        for( s= &(lattice[i]) ; i<end; i++,s++ ){
prefetch_matrix( &(s->link[dir]) );
prefetch_hwvec( (half_wilson_vector *)(gen_pt[0][i+1]) );
	    mult_su3_mat_hwvec( &(s->link[dir]), 
		(half_wilson_vector * )(gen_pt[0][i]), &hwv ); 
	    wp_grow( &hwv, (wilson_vector *)F_PT(s,dest), dir, isign);
	}
	mult_su3_mat_hwvec( &(s->link[dir]), 
	    (half_wilson_vector * )(gen_pt[0][i]), &hwv ); 
	wp_grow( &hwv, (wilson_vector *)F_PT(s,dest), dir, isign);
    }
    /* Don't do next for dir=TUP and isign=PLUS, i.e. in dslash! */
    else if(dir<TUP || isign==MINUS){
        for( s= &(lattice[i]) ; i<end; i++,s++ ){
prefetch_matrix( &(s->link[dir]) );
prefetch_wvec( (wilson_vector *)(F_PT((s+1),dest)) );
prefetch_hwvec( (half_wilson_vector *)(gen_pt[0][i+1]) );
	    mult_su3_mat_hwvec( &(s->link[dir]), 
	        (half_wilson_vector * )(gen_pt[0][i]), &hwv ); 
	    wp_grow_add( &hwv, (wilson_vector *)F_PT(s,dest), dir, isign);
        }
        mult_su3_mat_hwvec( &(s->link[dir]), 
	    (half_wilson_vector * )(gen_pt[0][i]), &hwv ); 
        wp_grow_add( &hwv, (wilson_vector *)F_PT(s,dest), dir, isign);
    }
#endif
	cleanup_gather(tag[0]);

        /* Take Wilson projection for src displaced in down direction,
        expand it, and add to dest */
	wait_gather(tag[1]);

    /* Don't do next for dir=TUP and isign=MINUS, i.e. in dslash^dagger! */
    if(dir<TUP || isign==PLUS){
#ifndef T3E
        FORSOMEPARITY(i,s,parity){
	    wp_grow_add( (half_wilson_vector *)(gen_pt[1][i]),
		(wilson_vector *)F_PT(s,dest), dir, -isign);
        } END_LOOP
#else
    switch(parity){
	case EVEN:
	    i=0; end = even_sites_on_node; break;
	case ODD:
	    i = even_sites_on_node; end = sites_on_node; break;
	case EVENANDODD:
	    i=0; end = sites_on_node; break;
    }
    end--;
    for( s= &(lattice[i]) ; i<end; i++,s++ ){
prefetch_wvec( (wilson_vector *)(F_PT((s+1),dest)) );
prefetch_hwvec( (half_wilson_vector *)(gen_pt[1][i+1]) );
	    wp_grow_add( (half_wilson_vector *)(gen_pt[1][i]),
		(wilson_vector *)F_PT(s,dest), dir, -isign);
    }
    wp_grow_add( (half_wilson_vector *)(gen_pt[1][i]),
	(wilson_vector *)F_PT(s,dest), dir, -isign);
#endif
    } /* end if(dir<TUP || isign==PLUS) */

	cleanup_gather(tag[1]);

    } /* end loop over directions */
} /* end (of dslash_w.c) */


